
'''
utils.py: helper functions for DLP api

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


'''


from som.logman import bot

import os
import sys


def paginate_items(items,size=100):
    '''paginate_items will return a list of lists, each of a particular max
    size
    '''
    groups = []
    for idx in range(0, len(items), size):
        group = items[idx:idx+size]
        groups.append(group)
    return groups


def clean_text(text,findings):
    '''clean_text will remove phi findings from a text object
    :param text: the original text sent to the content.inspect DLP endpoint
    :param findings: the full response for the text.
    '''
    if 'findings' in findings:
        for finding in findings['findings']:
            label = "**%s**" %finding['infoType']['name']
            # Note sure if this is best strategy, we can start with it
            text = text.replace(finding['quote'],label)
    return text





#!/usr/bin/env python

'''
radiology.py: models specifically for radiology

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

from som.api.google.storage.client import ClientBase
from som.api.google.storage.models import *
from som.api.google.storage.utils import *
from som.logman import bot
from som.utils import read_json
from som.wordfish.structures import structure_dataset
import six

######################################################################################
# Specs and base data structures for each radiology model
######################################################################################


def entity(uid,collection,metadata=None):
    '''entity returns an entity object
    parent is a collection
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata}]

    if type(collection) not in six.string_types:
        collection = collection.get_name()
    model = {'fields':fields,
             'key':['Collection',collection,'Entity', uid]}

    return model



def collection(uid,metadata=None,**fields):
    '''entity returns an entity object
    parent is an owner
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata}]

    model = {'fields':fields,
             'key':['Collection', uid]}
    return model
    


def image(uid,entity,download,url,storage,metadata=None,**fields):
    '''image returns an image object. entity is the parent
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata},
              {'key':'download','required':True,'value':download,
               'key':'url','required':True,'value': url,
               'key':'storage','required':True,'value': storage}]

    collection = entity.collection.get_name()
    entity = entity.get_name()

    model = {'fields':fields,
             'exclude_from_indexes': ['storage','download','url'],
             'key':['Collection', collection, 'Entity',entity,'Image', uid]}

    return model
    


def text(uid,entity,content,metadata=None,**fields):
    '''text returns a text object. entity is the parent
    '''
    fields = [{'key':'uid','required':True,'value':uid},
              {'key':'metadata','required':False,'value':metadata},
              {'key':'content','required':True,'value':content}]

    collection = entity.collection.get_name()
    entity = entity.get_name()

    model =  {'fields':fields,
              'exclude_from_indexes': ['storage','download','url'],
              'key':['Collection', collection, 'Entity',entity,'Text', uid]}
    return model




######################################################################################
# Core Models, extending base model
######################################################################################


class Collection(ModelBase):
  
    def __init__(self,client,uid,create=True,**fields):
        self.model = collection(uid=uid,**fields)
        super(Collection, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)


class Entity(ModelBase):

    def __init__(self,client,collection,uid,create=True,**fields):
        self.collection = collection
        self.model = entity(uid=uid, collection=collection,**fields)
        super(Entity, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)
        

    def collection(self,client):
        '''collection will return the collection associated with
        the entity
        '''
        return client.get(collection.key)


    def images(self,client):
        '''images will return images associated with the entity
        '''
        key = self.get_keypath()
        return client.get(client.key(*key, "Image"))
    

    def text(self,client):
        '''text will return text associated with the entity
        '''
        key = self.get_keypath()
        return client.get(client.key(*key, "Text"))


class Image(ModelBase):

    def __init__(self,client,uid,entity,create=True,url=None,
                 download=None,storage=None,**fields):
        self.entity = entity
        self.model = image(uid=uid,entity=entity,url=url,
                           storage=storage,download=download,**fields)
        super(Image, self).__init__(client,**fields)
        if create:
            self.this = self.update_or_create(client)
        else:
            self.this = self.setup(client)

    def __repr__(self):
        return self.this


class Text(ModelBase):

    def __init__(self,client,uid,entity,content,create=True,**fields):
        self.entity = entity
        self.model = text(uid=uid,entity=entity,content=content,**fields)

        super(Text, self).__init__(client,**fields)
        if create:
            self.this = self.get_or_create(client)
        else:
            self.this = self.setup(client)


    def __repr__(self):
        return self.this




######################################################################################
# Radiology Client to interact with Models
######################################################################################


class Client(ClientBase):

    def __init__(self,**kwargs):
        self.bucket_name = 'radiology'
        self.batch = BatchManager()
        super(Client, self).__init__(**kwargs)
    
    def __str__(self):
        return "storage.google.%s" %self.bucket_name

    def __repr__(self):
        return "storage.google.%s" %self.bucket_name


    ###################################################################
    ## Create and get #################################################
    ###################################################################

    def get_collection(self,fields):
        return Collection(client=self.datastore,**fields)

    def get_entity(self,fields):
        return Entity(client=self.datastore,**fields)


    def create_image(self,fields):
        return Image(client=self.datastore,**fields)

    def create_text(self,fields):
        return Text(client=self.datastore,**fields)

    def get_storage_path(self,file_path,collection_name,return_folder=False):
        folder = file_path.split(collection_name)[1]
        bucket_path = "%s%s" %(collection_name,
                               folder)
        if return_folder:
            return os.path.dirname(bucket_path)
        return bucket_path



    ###################################################################
    ## Upload #########################################################
    ###################################################################

    def upload_texts(self,texts,entity,batch=True):
        '''upload_texts will add text objects to the batch manager'''

        if not isinstance(texts,list):
            texts = [texts]

        new_texts = []
        for txt in texts:
            uid = self.get_storage_path(txt['original'],entity.collection.get_name())

            with open(txt['original'],'r') as filey:
                content = filey.read()

            fields = {'uid':uid,
                      'entity':entity,
                      'content':content,
                      'create': not batch }

            if 'metadata' in txt:
                fields['metadata'] = read_json(txt['metadata'])
 
            new_text = self.create_text(fields) # Create if not doing batch
            if batch:
                self.batch.add(new_text)

            bot.logger.debug('TEXT: %s',new_text)
            new_texts.append(new_text)
        return new_texts


    def upload_images(self,images,entity,batch=True):
        '''upload_images will add image objects to the batch manager, or create
        one at a time
        '''
        if not isinstance(images,list):
            images = [images]

        new_images = []
        collection_name = entity.collection.get_name()

        for img in images:
            bucket_folder = self.get_storage_path(img['original'],collection_name)
            image_obj = self.upload_object(file_path=img['original'],
                                           bucket_folder=bucket_folder)

            url = self.get_storage_path(image_obj['name'],collection_name)

            fields = {'uid':image_obj['id'],
                      'entity':entity,
                      'url':url,
                      'storage':image_obj,
                      'download':image_obj['mediaLink'],
                      'create': not batch }

            new_image = self.create_image(fields)
            new_images.append(new_image)

            bot.logger.debug('IMAGE: %s',new_image)

            if batch:
                self.batch.add(new_image)

        return new_images


    def upload_dataset(self,structures,batch=True):
        '''upload takes a list of structures (eg from som.wordfish.structures
        and uploads to storage and datastore. 
        :param structures: the list of wordfish data structures
        :param batch: add entities in batches (recommended, default True)
        '''
        
        if not isinstance(structures,list):
            structures = [structures]

        # Each structure corresponds to a collection, a subset of entities
        for coll in structures:

            # The structure should be created first
            uid = os.path.basename(coll['collection']['name'])
            fields = {'uid': uid }
            if 'metadata' in coll['collection']:
                fields['metadata'] = read_json(coll['collection']['metadata'])        
            new_collection = self.get_collection(fields)
            bot.logger.debug('GET: %s',new_collection)

            # Add entities
            if 'entities' in coll['collection']:
                for ntt in coll['collection']['entities']:
                    
                    fields = {'uid': os.path.basename(ntt['entity']['id']),
                              'collection':new_collection }
                    entity = self.get_entity(fields)
                   
                    # Add images and text to entities
                    if 'texts' in ntt['entity']:
                        new_texts = self.upload_texts(texts=ntt['entity']['texts'],
                                                      entity=entity,
                                                      batch=batch)

                    if 'images' in ntt['entity']:
                        new_images = self.upload_images(images=ntt['entity']['images'],
                                                        entity=entity,
                                                        batch=batch)

                    # Run a transaction for put (insert) images and text, and clears queue
                    self.batch.insert(client=self.datastore)



'''
som.base: the base connection to the SOM API

'''

from som.logman import bot

from som.api.base.auth import (
    authenticate,
    refresh_access_token
)

from som.api import ApiConnection

api_base = "https://api.rit.stanford.edu/identifiers/api"
api_version = "v1"

class SomApiConnection(ApiConnection):

    def __init__(self, token=None, **kwargs):
        super(SomApiConnection, self).__init__(**kwargs)
 
        self.headers = None
        if token is None:
            self.token = authenticate()
        else:
            self.token = token
        self.update_headers()
        

    def call(self,url,func,data=None,return_json=True):
        '''call overrides post for the som api to add the option
        to refresh a token given a 401 response.
        :param url: the url to send file to
        :param data: additional data to add to the request
        :param return_json: return json if successful
        '''
        response = func(url,self.headers,json=data)

        # Errored response, try again with refresh
        if response.status_code == 401:
            bot.logger.warning("Expired token, refreshing...")
            self.token = refresh_access_token()
            response = func(url,self.headers,json=data)

        if response.status_code == 200 and return_json:
            return response.json()

        return response



'''

client.py: simple clients for som api

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from som.api.base.connect import (
    api_base,
    api_version,
    SomApiConnection
)

from som.api.base.standards import spec
from som.logman import bot

class ClientBase(object):


    def __init__(self, spec=None, token=None, base=None, version=None):
 
        # Base and version consistent across
        if base == None:
            base = api_base

        if version == None:
            version = api_version

        self.base = base
        self.version = version
        self.study = None

        if spec != None:
            self.spec = spec

        self.api = SomApiConnection(token)


    def get_base(self,study=None):
        '''get_base returns the base api for a particular study
        '''
        if study is None:
            study = self.study
        return "%s/%s/%s" %(self.base,self.version,study)


    def deidentifiy_update(self,identifiers,test=False):
        '''deidentify update calls deidentify, but setting save_records
        to True to ensure that data is updated/written.
        '''
        return self.deidentify(identifiers,test=test,save_records=True)


    def deidentify(self,ids,test=False,save_records=False):
        '''deidentify will take a list of identifiers, and return the deidentified.
        if save_records is True, will save records. If False (default) only
        returns identifiers.
        :param identifiers: a list of identifiers (e.g., people)
        '''    
        # Default is production endpoint particular to the study
        study = self.study

        # Saving records (uid) or not (mrn) changes the endpoint
        action = "mrn"
        if save_records == True:
            bot.logger.debug("save_records is %s, no new data will be saved.")
            action = "uid"
        else:
            bot.logger.debug("save_records is %s, data will be saved.")

        # Testing overrides all other specifications
        if test == True:
            study = 'test'

        url = "%s/%s/%s/%s" %(api_base,api_version,action,study)

        response = self.api.post(url,data=ids)
        if "results" in response:
            return response['results']
        return response



'''

logman.py: Python logger base for som-tools

Copyright (c) 2017 Vanessa Sochat

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


import logging
import os
import sys

class Logman:

    def __init__(self,stream=True,MESSAGELEVEL=None):
        self.level = get_logging_level(MESSAGELEVEL)
        if stream == True:
            logging.basicConfig(stream=sys.stdout,level=self.level)
        else:
            logging.basicConfig(level=self.level)
        self.logger = logging.getLogger('som')
        self.formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def get_logging_level(MESSAGELEVEL=None):
    '''get_logging_level will return a logging level based on first
    a variable going into the function, then an environment variable
    MESSAGELEVEL, and then the default is DEBUG.
    :param MESSAGELEVEL: the level to get.
    '''
    if MESSAGELEVEL == None:
        MESSAGELEVEL = os.environ.get("MESSAGELEVEL","DEBUG")

    if MESSAGELEVEL in ["DEBUG","INFO"]:
        print("Environment message level found to be %s" %MESSAGELEVEL)

    if MESSAGELEVEL == "FATAL":
        return logging.FATAL

    elif MESSAGELEVEL == "CRITICAL":
        return logging.CRITICAL

    elif MESSAGELEVEL == "ERROR":
        return logging.ERROR

    elif MESSAGELEVEL == "WARNING":
        return logging.WARNING

    elif MESSAGELEVEL == "INFO":
        return logging.INFO

    elif MESSAGELEVEL in "DEBUG":
        return logging.DEBUG

    return logging.DEBUG

bot = Logman()
